-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_jh_package.all;
use work.sha3_pkg.all;

entity jh_datapath_mem_ppl is	
	generic ( ppl : integer := 2; HS : integer := 256; UF : integer := 1 );
	port (
		-- external
		rst     : in std_logic;
		clk     : in std_logic;
		din     : in std_logic_vector(ppl*w-1 downto 0);	
		dout    : out std_logic_vector(w-1 downto 0);	
		
        --fsm 1		
		ein     : in std_logic_vector(ppl-1 downto 0);	
        ec      : in std_logic_vector(ppl-1 downto 0);	
        lc      : in std_logic_vector(ppl-1 downto 0);	
		zc0     : out std_logic_vector(ppl-1 downto 0);		   

		--Round		  		        
		emsg    : in std_logic_vector(ppl-1 downto 0);
		schain      : in std_logic;
		round       : in std_logic_vector(5-(UF-1) downto 0);
		smsg        : in std_logic_vector(log2(ppl)-1 downto 0);
		er          : in std_logic;
		sf          : in std_logic;
		
		-- Output				
		shift_dout  : in std_logic_vector(ppl-1 downto 0);
        load_id     : in std_logic_vector(ppl-1 downto 0);
        load_hash   : in std_logic_vector(ppl-1 downto 0);
		sout        : in std_logic_vector(log2(ppl)-1 downto 0);
		sid         : in std_logic
		
	);				  
end jh_datapath_mem_ppl;


architecture struct of jh_datapath_mem_ppl is 
	-- type declaration
	type block_array_type is array (0 to ppl-1) of std_logic_vector(b-1 downto 0);
	type mw_array_type is array ( 0 to ppl-1 ) of std_logic_vector(mw-1 downto 0);
	type c_array_type is array ( 0 to ppl-1) of std_logic_vector(15 downto 0);
		
	--input		  
	signal min, mreg : mw_array_type;
	constant iv : std_logic_vector(b-1 downto 0) := get_iv( HS );
	-- round constant
	type crd_out_ux_type is array (0 to UF-1) of std_logic_vector(crw-1 downto 0);
	signal crd_out, crd_out_pre : crd_out_ux_type;
	
	-- round		
	type rd_out_ux_type is array (0 to UF-1) of std_logic_vector(b-1 downto 0);
	signal rd_out : rd_out_ux_type;
	signal rd_out0_regged : std_logic_vector(b-1 downto 0);
	signal round_reg, round_reg2 : std_logic_vector(5-(UF-1) downto 0);
	
	signal mreg_sel, min_sel : std_logic_vector(mw-1 downto 0);
	signal grouped, dg : std_logic_vector(b-1 downto 0);
	signal rin, rout, rdreg : std_logic_vector(b-1 downto 0);
	signal hp, hm : std_logic_vector(b-1 downto 0);
	signal hp_or_iv : std_logic_vector(b-1 downto 0);
	signal c : c_array_type;
	
	-- id, hash		   
	constant id_size : integer := 16;
	signal id_in : std_logic_vector(ppl*id_size-1 downto 0);
	
	
	-- debug  						   
	-- signal crdo : std_logic_vector(b-1 downto 0);
	-- constant zeros : std_logic_vector(b-1-crw downto 0) := (others => '0');
	-- signal rout_m, rd_out_m,crd_out_m, rdreg_m, rin_m : std_logic_matrix;
	
begin	
	-- ///////////////////////////////////////////////////////
	-- FSM1 SECTION
	fsm1_section : for i in 0 to ppl-1 generate
		decounter_gen : decountern 
			generic map ( N => 16, sub => 1 ) 
			port map ( clk => clk, rst => '0', load => lc(i), en => ec(i), input => din(ppl*w-(i+1)*w+log2mw+15 downto ppl*w-(i+1)*w+log2mw), output => c(i));
		zc0(i) <= '1' when c(i) = 0 else '0';																																										
		shfin_gen : sipo 
			generic map ( N => mw,M => w) 
			port map (clk => clk, en => ein(i), input => din(ppl*w-i*w-1 downto ppl*w-(i+1)*w), output => min(i) );
		min_reg : regn 
			generic map ( N => mw, init => mwzeros ) 
			port map ( clk => clk, rst => '0', en => emsg(i), input => min_sel, output => mreg(i) );
	end generate;

	-- ///////////////////////////////////////////////////////
	-- FSM2 SECTION
	-- Message selectors	 
	ppl2 : if ppl = 2 generate
		min_sel <= min(0) when smsg(0) = '0' else min(1);
		mreg_sel <= mreg(0) when smsg(0) = '0' else mreg(1);
	end generate;	   
	
	ppl4 : if ppl = 4 generate   
		with smsg(1 downto 0) select		
		min_sel <= min(0) when "00", min(1) when "01", min(2) when "10", min(3) when others;
		with smsg(1 downto 0) select		
		mreg_sel <= mreg(0) when "00", mreg(1) when "01", mreg(2) when "10", mreg(3) when others;
	end generate;	   
	
	-- input to r reg
	hm <= ( min_sel xor hp_or_iv(b-1 downto b/2) ) & hp_or_iv(b/2-1 downto 0);	
	hp_or_iv <= iv when sf = '1' else hp;	
	
	-- group (rearrange them into correct order)
	grouped <= form_group( hm, b, crw );			

	rin <= grouped when schain = '1' else rd_out(UF-1);

	-- Input Register to Round
	rreg_gen : regn 
		generic map ( N => b, init => bzeros ) 
		port map ( clk => clk, rst => '0', en => er, input => rin, output => rout );		
			
	--- RC	
	process ( clk )         
	begin
		if rising_edge( clk ) then
			crd_out <= crd_out_pre;	  			
		end if;
	end process;
	
	
	-- ROUND			
	UnrolledX1Gen: 
		if UF = 1 generate
			crd_out_pre(0) <= rc_cons( conv_integer(round));
			rd_gen 	: entity work.jh_rd_ppl(struct) 
				generic map ( bw => b, cw => crw  ) 
				port map ( clk => clk, input => rout, cr => crd_out(0), output => rd_out(0) ); 
		end generate;
	UnrolledX2Gen: 
		if UF = 2 generate					  
			crd_out_pre(0) <= rc_cons_half(0, conv_integer(round));
			
			
			stage0_reg : process ( clk ) 
			begin			   
				if rising_edge( clk ) then
					rd_out0_regged <= rd_out(0);  
					round_reg <= round;	 
					round_reg2 <= round_reg;
				end if;
			end process;
			
			ppl2 : if ppl = 2 generate 		  
				crd_out_pre(1) <= rc_cons_half(1, conv_integer(round_reg));
				rd_gen1 	: entity work.jh_rd(struct) 
					generic map ( bw => b, cw => crw  ) 
					port map ( input => rout, cr => crd_out(0), output => rd_out(0) ); 	
				rd_gen2 	: entity work.jh_rd(struct) 
					generic map ( bw => b, cw => crw  ) 
					port map ( input => rd_out0_regged, cr => crd_out(1), output => rd_out(1) ); 
			end generate;
			ppl4 : if ppl = 4 generate	  
				crd_out_pre(1) <= rc_cons_half(1, conv_integer(round_reg2));
				rd_gen1 	: entity work.jh_rd_ppl(struct) 
					generic map ( bw => b, cw => crw  ) 
					port map ( clk => clk, input => rout, cr => crd_out(0), output => rd_out(0) ); 
				rd_gen2 	: entity work.jh_rd_ppl(struct) 
					generic map ( bw => b, cw => crw  ) 
					port map ( clk => clk, input => rd_out0_regged, cr => crd_out(1), output => rd_out(1) ); 
			end generate;			
		end generate;			
		
	rdreg_gen : regn 
		generic map ( N => b, init => bzeros ) 
		port map ( clk => clk, rst => '0', en => er, input => rd_out(UF-1), output => rdreg );		
	
	dg <= degroup( rdreg, b, crw );
 	hp <= dg(b-1 downto b/2) & ( mreg_sel xor dg(b/2-1 downto 0) );	 
	-- ///////////////////////////////////////////////////////
	-- FSM3 SECTION
	-- Round Finalization / Load new block / Output generation
	fsm3_output : entity work.ppl_output_unit(struct) 
		generic map ( h => HS, w => w, ppl => ppl ) 
		port map ( clk => clk, rst => rst,
					hash_result => hp(HS-1 downto 0), id_in => id_in,
					sid => sid, shift_dout => shift_dout, load_hash => load_hash, 
					load_id => load_id, sout => sout,
					dout => dout );			 
			
	id_in_gen : for i in 0 to ppl-1 generate
		id_in((ppl-i)*id_size-1 downto (ppl-i-1)*id_size) <= din(ppl*w-(i+1)*w+15 downto ppl*w-(i+1)*w);
	end generate; 		

--	--debug
	 --rin_m <= blk2wordmatrix_inv( rin );
--	  rout_m <= blk2wordmatrix_inv( rout );
--	  rd_out_m <= blk2wordmatrix_inv( rd_out(0) );
--	  rdreg_m <= blk2wordmatrix_inv( rdreg );
--	dg_m <= blk2wordmatrix_inv( dg );
--	hp_m <= blk2wordmatrix_inv( hp );
--	hm_m <= blk2wordmatrix_inv( hm );
--	crdo <= crd_out(ux-1) & zeros;
--	crd_out_m <= blk2wordmatrix_inv( crdo );
end struct;		   		   